# Copyright (c) Alibaba, Inc. and its affiliates.
context_template = [
    '{}',
    'Human: {}',
    'Input: {}',
    'User: {}',
    'The city is beautiful, {}',
    'Today is a sunny day, {}',
    'America has a long coastline. {}',
    'The dogs are barking. {}',
    'These flowers need water. {}',
    'This city is good for the health, {}',
    'They are good at cooking fish and noodles, {}',
    'The supermarket here sells cheap today, {}',
    '今天是个晴天，{}',
    '这座城市很漂亮，{}',
    '获取更多信息，{}',
    '假设你是个人工智能小助手，{}',
    '这是个宝藏博主。{}',
    '北京是中国的首都，{}',
    '获得更多信息请点击相应的信息。{}',
    '三峡大坝是个伟大的建筑。{}',
]
